#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/* Struct of a generic heap.
 * @p	-	keeps the pointers
 * to the objects in the heap
 * @s	-	is the number of elements
 * in the heap.
 * @cmp	-	is a function pointer that
 * compares 2 elements from the heap.
 * */
#if __STDC__
#define inline
#endif

#ifndef HEAP_MAX_ELEMETS_TYPE
#define HEAP_MAX_ELEMENTS_TYPE	size_t
#endif

#ifndef HEAP_RETURN_TYPE
#define HEAP_RETURN_TYPE		char
#endif

/* Heap Structure where:
 * @p	-	Vector that points to data.
 * @max	-	Is the maximum number of elements the heap can take
 * 			before resizing the @p vector.
 * @s	-	Number of elements in the heap structure.
 */
struct heap{
	void **p;
	HEAP_MAX_ELEMENTS_TYPE s;
	HEAP_RETURN_TYPE (*cmp)( const void *v1, const void *v2 );
};

/* Initializes a heap structure, this function does
 * not allocate it.
 * @h	-	a struct heap pointer
 * @cmp	-	the compare function mentioned above,
 * at the struct discription
 */
static inline void heap_init( struct heap *h,
	   	HEAP_RETURN_TYPE (*cmp)( const void *v1, const void *v2 ) ){
	h->p = NULL;
	h->s = 0;
	h->cmp = cmp;
}

/* Internal usage function, you shoudn't use it whithout knowing what
 * you are doing.
 */

/* try to swap element i with his parent
 * @h	-	heap structure
 * $i	-	index of element in the heap
 */
static void __heap_up( struct heap *h, size_t i ){
	size_t j;
	void *aux;

	/* will get nasty if unsigned goes < 0 */
	if( i == 0 )
		return;
	j = ( i )/2;
	if( h->cmp( h->p[i], h->p[j] ) > 0 ){
		aux = h->p[i];
		h->p[i] = h->p[j];
		h->p[j] = aux;
		__heap_up( h, j );
	}
}

/* Internal usage function, you shoudn't use it whithout knowing what
 * you are doing.
 */
/* try to swap element i with one of it`s sons
 * @h	-	heap structure
 * $i	-	index of element in the heap
 */
static void __heap_down( struct heap *h, size_t i ){
	size_t j;
	void *aux;

	if( h->s < 1 )
		return;
	j = ( 2*i ) +1;
	if( j <= h->s ){
		if( j < h->s )
			if( h->cmp( h->p[j+1], h->p[j]) > 0 )
				j++;
		if( h->cmp( h->p[i], h->p[j] ) <= 0 ){
			aux = h->p[i];
			h->p[i] = h->p[j];
			h->p[j] = aux;
			__heap_down( h, j );
		}
	}
}

/* return the number of elements in the heap
 * @h	-	heap structure.
 */
static inline size_t heap_size( struct heap *h ){
	return h->s;
}

/* return true if heap is empty ( is kind of redundant )
 * @h	-	heap structure.
 */
static inline char heap_empty( struct heap *h ){
	return h->s ? 0:1;
}

/* UNIMPLEMENTED */
/* reserve n elements in the heap if n is bigger then the number
 * of elements in use.
 * @h	-	heap sructure.
 * @n	-	new number of elements
 */

/*
   static inline void heap_reserve( struct heap *h, size_t n ){
   if( n > h->s ){
   h->max = n;
   h->p = realloc( h->p, sizeof(h->p)*h->max );
   }
   }
   */

/* append an element to the heap.
 * @h	-	heap structure
 * @el	-	element pointer
 */
static inline void heap_push( struct heap *h, void *el ){

	h->s++; 
	h->p = realloc( h->p, sizeof(*(h->p))*h->s );
	assert( h->p != NULL );
	h->p[ h->s -1 ] = el;
	__heap_up( h, h->s - 1 );
}

/* return element from the heap
 * @h	-	heap structure
 */
static inline void *heap_pop( struct heap *h ){

	void *ret;
	if( h->s == 0 )
		return NULL;
	ret = h->p[0];
	h->s--;
	h->p[0] = h->p[ h->s];
	h->p = realloc( h->p, sizeof(*(h->p))*h->s+1 );
	__heap_down( h, 0 );
	return ret;
}

/* dumps the elements from the heap to a
 * determined function, for testing.
 * @h	-	heap structure
 * @dump-	function that will recieve the elemets.
 */
static inline void heap_dump( struct heap *h, void (*dump)( const void*) ){
	size_t i;
	printf("max=<");
	dump( h->p[0] );
	printf(">:\n");
	for( i=0; i < h->s; i++ )
		dump( h->p[i] );
	printf("\n");
}

/* Example of usage:
 *
 * first you need a function that will compare
 * witch of 2 given elements is bigger like in qsort
 *
HEAP_RETURN_TYPE cmp( const void *v1, const void *v2 ){
	return *( (int*)v1 ) - *( (int*)v2 );
}
*/

/* Here is the main, you can see where to allocate
 * the memory for the elements and where to free
 * them.
 *

int main(){
	struct heap h;
	int n,i;
	heap_init( &h, cmp );

	while( scanf(" %d", &n ) != EOF ){
		int *k = malloc( sizeof(int) );
		*k = n;
		heap_push( &h, k );
	}
	for( i=0; i < n; i++ ){
		int *t;
		t = heap_pop( &h );
		printf("%d ", *t );
		free( t );
	}
	printf("\n");
	return 0;
}

*/

